# BitIO is available at http://github.com/naoya/perl-BitIO/tree/master
package HuffmanCoder;
use strict;
use warnings;

use Params::Validate qw/validate_pos ARRAYREF/;
use Heap::Simple;
use BitIO;

sub new {
    my ($class, $count) = (@_, 1, { type => ARRAYREF });
    return bless {
        count => $count,
        root  => build_huffman_tree($count),
    }
}

sub build_huffman_tree {
    my $count = shift;
    my $heap = Heap::Simple->new(elements => 'Array', order => '<');

    for (my $i = 0; $i < @$count; $i++) {
        if ($count->[$i] > 0) {
            $heap->insert([ $count->[$i], { symbol => $i } ]);
        }
    }

    while (1) {
        my $x = $heap->extract_first;
        my $y = $heap->extract_first;

        if (not defined $y) {
            return $x->[1];
        }

        my $z = {};
        $z->{left}  = $x->[1];
        $z->{right} = $y->[1];
        $heap->insert([ $x->[0] + $y->[0], $z ]);
    }
}

sub show_code {
    my $self = shift;
    _show_code($self->{root}, []);
}

sub _show_code {
    my ($node, $bits) = @_;
    if (defined $node->{symbol}) {
        printf "%s => %s\n", chr $node->{symbol}, join('', @$bits);
    } else {
        my @left = @$bits;
        push @left, 0;
        _show_code($node->{left},  \@left);

        my @right = @$bits;
        push @right, 1;
        _show_code($node->{right}, \@right);
    }
}

sub encode {
    my ($self, $in, $out, $size) = @_;

    my $code_table = [];
    _make_code($code_table, $self->{root}, 0, 0);

    my $writer = BitIO->new($out);
    $writer->putbits(32, $size);

    my $cnt_bin = pack('w*', @{$self->{count}});
    $writer->putbits(32, length $cnt_bin);
    $writer->write($cnt_bin) or die $!;

    while (defined(my $c = $in->getc)) {
        my ($n, $code) = @{$code_table->[ord $c]};
        $writer->putbits($n, $code);
    }

    $writer->flush;
}

sub _make_code {
    my ($table, $node, $n, $code) = @_;

    if (defined $node->{symbol}) {
        $table->[$node->{symbol}] = [$n, $code];
    } else {
        _make_code($table, $node->{left},  $n + 1, $code << 1);
        _make_code($table, $node->{right}, $n + 1, ($code << 1) | 1);
    }
}

sub decode {
    my ($class, $in, $out) = @_;
    my $reader = BitIO->new($in);

    my $size     = $reader->getbits(32);
    my $cnt_size = $reader->getbits(32);
    $reader->read(\my $cnt_bin, $cnt_size) or die $!;
    my @count = unpack('w*', $cnt_bin);

    my $huff = $class->new(\@count);
    while ($size > 0) {
        my $node = $huff->{root};
        while (not defined $node->{symbol}) {
            $node = $reader->getbit == 0 ? $node->{left} : $node->{right};
        }
        $out->print(chr $node->{symbol});
        $size--;
    }
}

1;
